from django.contrib import admin
from django.urls import include, path, URLPattern, URLResolver
from rest_framework.authtoken.views import obtain_auth_token
from rest_framework.routers import DefaultRouter

from cardroom.apps import CardroomConfig
from cardroom.utilities import get_admin, get_auth, get_felt
from cardroom.views import (
    CashGameFeltView,
    CashGameFrameView,
    CashGameViewSet,
    HandHistoryFeltView,
    HandHistoryFramesView,
    HandHistoryViewSet,
    PokerViewSet,
    TableViewSet,
)

app_name: str = CardroomConfig.name
router: DefaultRouter = DefaultRouter()
urlpatterns: list[URLPattern | URLResolver] = [
    path(
        'cash-games/<int:pk>/frame/',
        CashGameFrameView.as_view(),
        name='cashgame_frame',
    ),
    path(
        'hand-histories/<int:pk>/frames/',
        HandHistoryFramesView.as_view(),
        name='handhistory_frames',
    ),
]

router.register('poker', PokerViewSet)
router.register('tables', TableViewSet)
router.register('cash-games', CashGameViewSet)
router.register('hand-histories', HandHistoryViewSet)

if get_auth():
    urlpatterns.append(path('auth/', include('rest_framework.urls')))
    urlpatterns.append(path('auth/token/', obtain_auth_token))

if get_admin():
    urlpatterns.append(path('admin/', admin.site.urls))

if get_felt():
    urlpatterns.append(
        path(
            'cash-games/<int:pk>/felt/',
            CashGameFeltView.as_view(),
            name='cashgame_felt',
        ),
    )
    urlpatterns.append(
        path(
            'hand-histories/<int:pk>/felt/',
            HandHistoryFeltView.as_view(),
            name='handhistory_felt',
        ),
    )

urlpatterns.append(path('', include(router.urls)))



from abc import ABC, abstractmethod
from threading import Thread, Lock
from time import sleep

from asgiref.sync import async_to_sync
from channels.generic.websocket import JsonWebsocketConsumer
from channels.layers import get_channel_layer
from django.db import OperationalError, ProgrammingError

from cardroom.models import CashGame
from cardroom.utilities import get_gamemaster_timeout, serialize

controllers_lock = Lock()
controllers = {}
threads = {}


def set_controller(model):
    controllers[model.group_name] = model.load()
    threads[model.group_name] = Thread(
        target=controllers[model.group_name].mainloop,
        daemon=True,
    )

    threads[model.group_name].start()


def get_controller(model):
    with controllers_lock:
        return controllers[model.group_name]


def handle(model, user, content):
    get_controller(model).handle(user.username, content)


frames_lock = Lock()
frames = {}


def broadcast(model, sub_frames, users_message):
    if sub_frames is not None and sub_frames:
        with frames_lock:
            frames[model.group_name] = sub_frames[-1]

        async_to_sync(get_channel_layer().group_send)(
            model.group_name,
            {'type': 'update', 'frames': serialize(sub_frames)},
        )

    if all(users_message):
        users, message = users_message

        async_to_sync(get_channel_layer().group_send)(
            model.group_name,
            {'type': 'notify', 'users': users, 'message': message},
        )


def get_frames(model):
    with frames_lock:
        return frames[model.group_name]


class ControllerConsumer(JsonWebsocketConsumer, ABC):
    @property
    def pk(self):
        return self.scope['url_route']['kwargs']['pk']

    @property
    def user(self):
        return self.scope['user']

    @property
    @abstractmethod
    def controller(self):
        pass

    def connect(self):
        super().connect()
        async_to_sync(self.channel_layer.group_add)(
            self.controller.group_name,
            self.channel_name,
        )
        self.update(
            {
                'type': 'update',
                'frames': [serialize(get_frames(self.controller))],
            },
        )

    def disconnect(self, code):
        async_to_sync(self.channel_layer.group_discard)(
            self.controller.group_name,
            self.channel_name,
        )
        super().disconnect(code)

    def receive_json(self, content, **kwargs):
        if self.user.is_authenticated:
            handle(self.controller, self.user, content)
        else:
            self.notify(
                {
                    'type': 'notify',
                    'users': [''],
                    'message': 'you are unauthenticated',
                },
            )

    def update(self, event):
        frames = []

        for sub_frames in event['frames']:
            frames.append(sub_frames.get(self.user.username, sub_frames['']))

        self.send_json(event | {'frames': frames})

    def notify(self, event):
        if self.user.username in event['users']:
            self.send_json(event)


class CashGameConsumer(ControllerConsumer):
    @property
    def controller(self):
        return CashGame.objects.get(pk=self.pk)


def mainloop():  # TODO: revert to signal based approach
    while True:
        try:
            for cash_game in CashGame.objects.all():
                try:
                    get_controller(cash_game)
                except KeyError:
                    set_controller(cash_game)
        except (OperationalError, ProgrammingError):
            pass

        sleep(get_gamemaster_timeout())


MAINLOOP_THREAD = Thread(target=mainloop)

MAINLOOP_THREAD.start()



from pickle import dumps
from textwrap import dedent

from django.test import TestCase
from pokerkit import Automation, NoLimitShortDeckHoldem, NoLimitTexasHoldem
import pokerkit

from cardroom.models import Poker, Table, HandHistory
from cardroom.utilities import get_divmod
import cardroom.table as table


class PickleTestCaseMixin(TestCase):
    def assertPickleEqual(self, o1: object, o2: object) -> None:
        self.assertEqual(dumps(o1), dumps(o2))


class PokerTestCase(PickleTestCaseMixin, TestCase):
    def test_load(self) -> None:
        self.assertPickleEqual(
            Poker.objects.create(
                variant='NT',
                raw_antes={1: 3},
                raw_blinds_or_straddles=[1, 2],
                min_bet=2,
            ).load(),
            NoLimitTexasHoldem(
                Poker.automations,
                False,
                {1: 3},
                [1, 2],
                2,
                get_divmod(),
            ),
        )
        self.assertPickleEqual(
            Poker.objects.create(
                variant='NT',
                raw_antes=[0, 3],
                raw_blinds_or_straddles=[1, 2],
                min_bet=2,
            ).load(),
            NoLimitTexasHoldem(
                Poker.automations,
                False,
                [0, 3],
                [1, 2],
                2,
                get_divmod(),
            ),
        )
        self.assertPickleEqual(
            Poker.objects.create(
                variant='NT',
                ante_trimming_status=True,
                raw_antes=0,
                raw_blinds_or_straddles=[1, 2],
                min_bet=2,
            ).load(),
            NoLimitTexasHoldem(
                Poker.automations,
                True,
                0,
                [1, 2],
                2,
                get_divmod(),
            ),
        )
        self.assertPickleEqual(
            Poker.objects.create(
                variant='NS',
                ante_trimming_status=True,
                raw_antes=2,
                raw_blinds_or_straddles={-1: 2},
                min_bet=2,
            ).load(),
            NoLimitShortDeckHoldem(
                Poker.automations,
                True,
                2,
                {-1: 2},
                2,
                get_divmod(),
            ),
        )


class TableTestCase(PickleTestCaseMixin, TestCase):
    def test_load(self) -> None:
        game = Poker.objects.create(
            variant='NT',
            raw_antes={1: 3},
            raw_blinds_or_straddles=[1, 2],
            min_bet=2,
        )

        self.assertPickleEqual(
            Table.objects.create(
                game=game,
                seat_count=6,
                min_starting_stack=80,
                max_starting_stack=200,
            ).load(),
            table.Table(
                NoLimitTexasHoldem(
                    Poker.automations,
                    False,
                    {1: 3},
                    [1, 2],
                    2,
                    get_divmod(),
                ),
                6,
                80,
                200,
            ),
        )


class HandHistoryTestCase(TestCase):
    def test_dump_and_load(self) -> None:
        s = dedent(
            '''\
            # An example badugi hand from Wikipedia.
            # Link: https://en.wikipedia.org/wiki/Badugi

            variant = "FB"
            ante_trimming_status = true
            antes = [0, 0, 0, 0]
            blinds_or_straddles = [1, 2, 0, 0]
            small_bet = 2
            big_bet = 4
            starting_stacks = [200, 200, 200, 200]
            actions = [
              # Pre-draw

              "d dh p1 ????????",  # Bob
              "d dh p2 ????????",  # Carol
              "d dh p3 ????????",  # Ted
              "d dh p4 ????????",  # Alice

              "p3 f",  # Ted
              "p4 cc",  # Alice
              "p1 cc",  # Bob
              "p2 cc",  # Carol

              # First draw

              "p1 sd ????",  # Bob
              "p2 sd ????",  # Carol
              "p4 sd ??",  # Alice
              "d dh p1 ????",  # Bob
              "d dh p2 ????",  # Carol
              "d dh p4 ??",  # Alice

              "p1 cc",  # Bob
              "p2 cbr 2",  # Carol
              "p4 cc",  # Alice
              "p1 cc",  # Bob

              # Second draw

              "p1 sd ??",  # Bob
              "p2 sd",  # Carol
              "p4 sd ??",  # Alice
              "d dh p1 ??",  # Bob
              "d dh p4 ??",  # Alice

              "p1 cc",  # Bob
              "p2 cbr 4",  # Carol
              "p4 cbr 8",  # Alice
              "p1 f",  # Bob
              "p2 cc",  # Bob

              # Third draw

              "p2 sd ??",  # Carol
              "p4 sd",  # Alice
              "d dh p2 ??",  # Carol

              "p2 cc",  # Carol
              "p4 cbr 4",  # Alice
              "p2 cc",  # Carol

              # Showdown

              "p4 sm 2s4c6d9h",  # Alice
              "p2 sm 3s5d7c8h",  # Carol
            ]
            author = "Juho Kim"
            players = ["Bob", "Carol", "Ted", "Alice"]
            '''
        )
        pkhh0 = pokerkit.HandHistory.loads(
            s,
            automations=(Automation.CARD_BURNING,),
        )
        hh0 = HandHistory.dump(pkhh0)
        hh0.save()
        hh1 = HandHistory.objects.first()

        assert hh1 is not None

        pkhh1 = hh1.load()

        self.assertEqual(pkhh0, pkhh1)



from __future__ import annotations

from collections.abc import Iterable, Iterator
from dataclasses import dataclass, field, KW_ONLY
from functools import partial
from itertools import chain
from typing import overload, TypeVar

from pokerkit import Card, HandHistory

from cardroom.table import Table

_T = TypeVar('_T')


@dataclass
class Configuration:
    _: KW_ONLY

    background_color: str

    table_border_color: str
    table_felt_color: str
    table_x: float
    table_y: float
    table_outer_width: float
    table_outer_height: float
    table_inner_width: float
    table_inner_height: float

    button_color: str
    button_angle: float
    button_ring_width: float
    button_ring_height: float
    button_diameter: float
    button_text_color: str
    button_text_style: str
    button_text_font: str
    button_text_size: float
    button_text: str

    board_x: float
    board_y: float
    board_width: float
    board_height: float
    board_radius: float
    board_color: str
    board_pot_text: str
    board_pot_text_color: str
    board_pot_text_style: str
    board_pot_text_font: str
    board_pot_text_size: float
    board_card_margin: float
    board_card_height: float
    board_card_radius: float
    board_card_color: str
    board_card_text_style: str
    board_card_text_font: str
    board_card_text_size: float

    bet_ring_width: float
    bet_ring_height: float
    bet_angle: float
    bet_box_color: str
    bet_box_x_padding: float
    bet_box_height: float
    bet_box_radius: float
    bet_text_style: str
    bet_text_color: str
    bet_text_font: str
    bet_text_size: float

    seat_ring_width: float
    seat_ring_height: float
    seat_angle: float

    hole_x: float
    hole_y: float
    hole_width: float
    hole_height: float
    hole_radius: float
    hole_color: str
    hole_card_margin: float
    hole_card_height: float
    hole_card_radius: float
    hole_card_color: str
    hole_card_text_style: str
    hole_card_text_font: str
    hole_card_text_size: float

    name_x: float
    name_y: float
    name_box_width: float
    name_box_height: float
    name_box_radius: float
    name_box_color: str
    name_text_style: str
    name_text_color: str
    name_text_font: str
    name_text_size: float

    stack_x: float
    stack_y: float
    stack_box_width: float
    stack_box_height: float
    stack_box_radius: float
    stack_box_color: str
    stack_text_style: str
    stack_text_color: str
    stack_text_font: str
    stack_text_size: float

    previous_action_x: float
    previous_action_y: float
    previous_action_box_width: float
    previous_action_box_height: float
    previous_action_box_radius: float
    previous_action_box_color: str
    previous_action_text_style: str
    previous_action_text_color: str
    previous_action_text_font: str
    previous_action_text_size: float

    club_color: str
    diamond_color: str
    heart_color: str
    spade_color: str
    unknown_color: str

    shifter_timeout: float
    watchdog_timeout: float

    frame_rate: float


@dataclass
class Frame:
    _: KW_ONLY
    names: list[str | None] = field(default_factory=list)
    button: int | None = None
    bets: list[int | None] = field(default_factory=list)
    stacks: list[int | None] = field(default_factory=list)
    pots: list[int] = field(default_factory=list)
    holes: list[list[Card] | None] = field(default_factory=list)
    hole_statuses: list[bool] = field(default_factory=list)
    board: list[Card] = field(default_factory=list)
    board_count: int = 0
    previous_action: tuple[int, str] | None = None
    actor: int | None = None
    join: list[int] | None = None
    leave: bool = False
    sit_out: bool = False
    be_back: bool = False
    buy_rebuy_top_off_or_rat_hole: tuple[int, int] | None = None
    stand_pat_or_discard: list[Card] | None = None
    fold: bool = False
    check_or_call: int | None = None
    post_bring_in: int | None = None
    complete_bet_or_raise_to: tuple[bool, bool, int, int] | None = None
    show_or_muck_hole_cards: bool | None = None

    @classmethod
    def from_table(cls, table: Table) -> dict[str, Frame]:
        button = None if table.button is None else table.button.seat_index
        hole_statuses = list[bool]()

        for street in table.game.streets:
            hole_statuses.extend(street.hole_dealing_statuses)

        if table.state is None:
            pots = []
            board = []
        else:
            pots = [pot.amount for pot in table.state.pots]
            board = table.state.board_cards.copy()

        board_count = table.game.max_board_card_count
        previous_action = None
        actor = None
        names = []
        bets = list[int | None]()
        stacks = list[int | None]()

        for seat in table.seats:
            names.append(seat.user)

            if seat.player_index is None or table.state is None:
                bets.append(None)
                stacks.append(seat.starting_stack)
            else:
                bets.append(table.state.bets[seat.player_index])
                stacks.append(table.state.stacks[seat.player_index])

                if (
                        (
                            table.state.stander_pat_or_discarder_index
                            == seat.player_index
                        )
                        or table.state.actor_index == seat.player_index
                        or table.state.showdown_index == seat.player_index
                ):
                    actor = seat.index

        frames = {}

        for user in chain(table.users, ('',)):
            holes = list[list[Card] | None]()

            if table.can_join(user):
                join = list(
                    filter(partial(table.can_join, user), table.seat_indices),
                )
            else:
                join = None

            leave = table.can_leave(user)
            sit_out = table.can_sit_out(user)
            be_back = table.can_be_back(user)

            if table.can_buy_rebuy_top_off_or_rat_hole(user):
                buy_rebuy_top_off_or_rat_hole = (
                    table.min_starting_stack,
                    table.max_starting_stack,
                )
            else:
                buy_rebuy_top_off_or_rat_hole = None

            stand_pat_or_discard = None
            fold = False
            check_or_call = None
            post_bring_in = None
            complete_bet_or_raise_to = None
            show_or_muck_hole_cards = None
            seat_or_none = table.get_seat_or_none(user)

            if seat_or_none is not None and seat_or_none is table.acting_seat:
                assert table.state is not None

                if table.state.can_stand_pat_or_discard():
                    assert (
                        table.state.stander_pat_or_discarder_index
                        is not None
                    )

                    stand_pat_or_discard = list(
                        table.state.get_down_cards(
                            table.state.stander_pat_or_discarder_index,
                        ),
                    )

                if table.state.can_fold():
                    fold = True

                if table.state.can_check_or_call():
                    check_or_call = table.state.checking_or_calling_amount

                if table.state.can_post_bring_in():
                    post_bring_in = table.state.bring_in

                if table.state.can_complete_bet_or_raise_to():
                    assert (
                        table.state.min_completion_betting_or_raising_to_amount
                        is not None
                    )
                    assert (
                        table.state.max_completion_betting_or_raising_to_amount
                        is not None
                    )

                    complete_bet_or_raise_to = (
                        table.state.completion_status,
                        any(table.state.bets),
                        (
                            table
                            .state
                            .min_completion_betting_or_raising_to_amount
                        ),
                        (
                            table
                            .state
                            .max_completion_betting_or_raising_to_amount
                        ),
                    )

                if table.state.can_show_or_muck_hole_cards():
                    assert table.state.showdown_index is not None

                    show_or_muck_hole_cards = table.state.can_win_now(
                        table.state.showdown_index,
                    )

            for seat in table.seats:
                if seat.player_status:
                    assert seat.player_index is not None
                    assert table.state is not None

                    hole_cards = list[Card]()

                    for card, status in zip(
                            table.state.hole_cards[seat.player_index],
                            table.state.hole_card_statuses[seat.player_index],
                    ):
                        if status or user == seat.user:
                            hole_cards.append(card)
                        else:
                            hole_cards.extend(Card.parse('??'))

                    holes.append(hole_cards)
                else:
                    holes.append(None)

            frames[user] = Frame(
                names=names,
                button=button,
                bets=bets,
                stacks=stacks,
                pots=pots,
                holes=holes,
                hole_statuses=hole_statuses,
                board=board,
                board_count=board_count,
                previous_action=previous_action,
                actor=actor,
                join=join,
                leave=leave,
                sit_out=sit_out,
                be_back=be_back,
                buy_rebuy_top_off_or_rat_hole=buy_rebuy_top_off_or_rat_hole,
                stand_pat_or_discard=stand_pat_or_discard,
                fold=fold,
                check_or_call=check_or_call,
                post_bring_in=post_bring_in,
                complete_bet_or_raise_to=complete_bet_or_raise_to,
                show_or_muck_hole_cards=show_or_muck_hole_cards,
            )

        return frames

    @classmethod
    def from_hand_history(cls, hand_history: HandHistory) -> Iterator[Frame]:

        @overload
        def p2s(players: int) -> int:
            ...

        @overload
        def p2s(players: Iterable[_T]) -> list[_T | None]:
            ...

        def p2s(players: int | Iterable[_T]) -> int | list[_T | None]:
            if isinstance(players, int):
                return indices[players]

            seats: list[_T | None] = [None] * count

            for i, player in enumerate(players):
                seats[indices[i]] = player

            return seats

        game = hand_history.create_game()

        if hand_history.seat_count is not None:
            count = hand_history.seat_count
        elif hand_history.seats is not None:
            count = max(hand_history.seats)
        else:
            count = len(hand_history.starting_stacks)

        if hand_history.seats is None:
            indices = list(range(count))
        else:
            indices = [seat - 1 for seat in hand_history.seats]

        if hand_history.players is None:
            names = p2s(f'p{i}' for i in range(1, count + 1))
        else:
            names = p2s(hand_history.players)

        if game.button_status:
            button = p2s(len(hand_history.starting_stacks) - 1)
        else:
            button = None

        for state, action in hand_history.iter_state_actions():
            bets = p2s(state.bets)

            if state.status or hand_history.finishing_stacks is None:
                stacks = p2s(state.stacks)
            else:
                stacks = p2s(hand_history.finishing_stacks)

            pots = [pot.amount for pot in state.pots]

            if not pots:
                pots.append(0)

            holes = p2s(map(list.copy, state.hole_cards))

            for i, hole in enumerate(holes):
                if not hole:
                    holes[i] = None

            hole_statuses = list[bool]()

            for street in game.streets:
                hole_statuses.extend(street.hole_dealing_statuses)

            board = state.board_cards.copy()
            board_count = game.max_board_card_count
            operation = state.operations[-1] if state.operations else None

            if (
                    action is not None
                    and operation is not None
                    and hasattr(operation, 'player_index')
            ):
                previous_action = p2s(operation.player_index), action
            else:
                previous_action = None

            if state.stander_pat_or_discarder_index is not None:
                actor = p2s(state.stander_pat_or_discarder_index)
            elif state.actor_index is not None:
                actor = p2s(state.actor_index)
            elif state.showdown_index is not None:
                actor = p2s(state.showdown_index)
            else:
                actor = None

            if state.can_stand_pat_or_discard():
                assert state.stander_pat_or_discarder_index is not None

                stand_pat_or_discard = list(
                    state.get_down_cards(state.stander_pat_or_discarder_index),
                )
            else:
                stand_pat_or_discard = None

            fold = state.can_fold()

            if state.can_check_or_call():
                check_or_call = state.checking_or_calling_amount
            else:
                check_or_call = None

            if state.can_post_bring_in():
                post_bring_in = state.bring_in
            else:
                post_bring_in = None

            if state.can_complete_bet_or_raise_to():
                assert (
                    state.min_completion_betting_or_raising_to_amount
                    is not None
                )
                assert (
                    state.max_completion_betting_or_raising_to_amount
                    is not None
                )

                complete_bet_or_raise_to = (
                    state.completion_status,
                    any(state.bets),
                    state.min_completion_betting_or_raising_to_amount,
                    state.max_completion_betting_or_raising_to_amount,
                )
            else:
                complete_bet_or_raise_to = None

            if state.can_show_or_muck_hole_cards():
                assert state.showdown_index is not None

                show_or_muck_hole_cards = state.can_win_now(
                    state.showdown_index,
                )
            else:
                show_or_muck_hole_cards = None

            yield Frame(
                names=names,
                button=button,
                bets=bets,
                stacks=stacks,
                pots=pots,
                holes=holes,
                hole_statuses=hole_statuses,
                board=board,
                board_count=board_count,
                previous_action=previous_action,
                actor=actor,
                stand_pat_or_discard=stand_pat_or_discard,
                fold=fold,
                check_or_call=check_or_call,
                post_bring_in=post_bring_in,
                complete_bet_or_raise_to=complete_bet_or_raise_to,
                show_or_muck_hole_cards=show_or_muck_hole_cards,
            )



from urllib.parse import parse_qs

from channels.db import database_sync_to_async
from django.contrib.auth.models import AnonymousUser
from rest_framework.authtoken.models import Token


@database_sync_to_async
def get_user_from_token(token_key):
    try:
        return Token.objects.get(key=token_key).user
    except Token.DoesNotExist:
        return AnonymousUser()


class TokenAuthMiddleware:
    """The Token auth middleware for Django Channels."""

    def __init__(self, inner):
        self.inner = inner

    async def __call__(self, scope, receive, send):
        query_string = parse_qs(scope['query_string'].decode())

        if 'token' in query_string:
            scope['user'] = await get_user_from_token(query_string['token'][0])
        elif 'user' not in scope:
            scope['user'] = AnonymousUser()

        return await self.inner(scope, receive, send)


def TokenAuthMiddlewareStack(inner):
    return TokenAuthMiddleware(inner)



from django.urls import path, URLResolver

from cardroom.apps import CardroomConfig
from cardroom.consumers import CashGameConsumer

app_name: str = CardroomConfig.name
urlpatterns: list[URLResolver] = [
    path(
        'cash-games/<int:pk>/',
        CashGameConsumer.as_asgi(),
        name='cashgame_websocket',
    ),
]



from argparse import ArgumentParser
from functools import partial
from glob import glob
from itertools import chain
from typing import Any

from django.core.management.base import BaseCommand, CommandError
from pokerkit import HandHistory

from cardroom import models


class Command(BaseCommand):
    help = 'Installs the hand histories to the database.'

    def add_arguments(self, parser: ArgumentParser) -> None:
        parser.add_argument(
            'pathnames',
            nargs='+',
            type=str,
            help='One or more pathnames',
        )

    def handle(self, *args: Any, **options: Any) -> None:
        pathnames = options['pathnames']
        count = 0

        for pathname in chain.from_iterable(
                map(partial(glob, recursive=True), pathnames),
        ):
            try:
                with open(pathname, 'rb') as file:
                    hh = HandHistory.load(file)
            except ValueError:
                raise CommandError(
                    f'failed to parse hand history file {repr(pathname)}',
                )

            models.HandHistory.dump(hh).save()

            count += 1

        self.stdout.write(
            self.style.SUCCESS(
                f'Success: created {count} hand history models',
            ),
        )



from importlib import import_module

from channels.auth import AuthMiddlewareStack
from channels.routing import ProtocolTypeRouter, URLRouter
from channels.security.websocket import AllowedHostsOriginValidator
from django.core.asgi import get_asgi_application

from cardroom.middleware import TokenAuthMiddlewareStack
from cardroom.utilities import get_root_routingconf

application = ProtocolTypeRouter(
    {
        'http': get_asgi_application(),
        'websocket': AllowedHostsOriginValidator(
            AuthMiddlewareStack(
                TokenAuthMiddlewareStack(
                    URLRouter(
                        import_module(get_root_routingconf()).urlpatterns,
                    ),
                ),
            ),
        )
    },
)



from collections.abc import Callable, Iterable, Mapping
from dataclasses import asdict, is_dataclass
from math import inf, floor, pi
from typing import Any, cast
from zoneinfo import ZoneInfo
import math

from django.conf import settings
from django.utils.module_loading import import_string
import pokerkit

from cardroom.felt import Configuration

DEFAULT_DIVMOD: str = 'cardroom.utilities.divmod'
DEFAULT_PARSE_VALUE: str = 'cardroom.utilities.parse_value'
DEFAULT_DECIMAL_PLACES: int | float = inf
DEFAULT_AUTH: bool = False
DEFAULT_ADMIN: bool = False
DEFAULT_FELT: bool = False
DEFAULT_CONFIGURATION: Configuration = Configuration(
    background_color='gray',
    table_border_color='saddlebrown',
    table_felt_color='seagreen',
    table_x=0,
    table_y=0,
    table_outer_width=0.6,
    table_outer_height=0.35,
    table_inner_width=0.55,
    table_inner_height=0.3,
    button_color='white',
    button_angle=pi / 24,
    button_ring_width=0.5,
    button_ring_height=0.25,
    button_diameter=0.035,
    button_text_color='black',
    button_text_style='bold',
    button_text_font='sans',
    button_text_size=0.025,
    button_text='B',
    board_x=0,
    board_y=-0.05,
    board_width=0.3,
    board_height=0.025,
    board_radius=0.0125,
    board_color='black',
    board_pot_text='Pot: ',
    board_pot_text_color='white',
    board_pot_text_style='bold',
    board_pot_text_font='sans',
    board_pot_text_size=0.02,
    board_card_margin=0.0025,
    board_card_height=0.06,
    board_card_radius=0.005,
    board_card_color='white',
    board_card_text_style='bold',
    board_card_text_font='sans',
    board_card_text_size=0.025,
    bet_ring_width=0.425,
    bet_ring_height=0.175,
    bet_angle=0,
    bet_box_color='black',
    bet_box_x_padding=0.025,
    bet_box_height=0.025,
    bet_box_radius=0.0125,
    bet_text_style='bold',
    bet_text_color='white',
    bet_text_font='sans',
    bet_text_size=0.02,
    seat_ring_width=0.8,
    seat_ring_height=0.55,
    seat_angle=0,
    hole_x=0,
    hole_y=0,
    hole_width=0.225,
    hole_height=0.03,
    hole_radius=0.005,
    hole_color='black',
    hole_card_margin=0.003,
    hole_card_height=0.05,
    hole_card_radius=0.004,
    hole_card_color='white',
    hole_card_text_style='bold',
    hole_card_text_font='sans',
    hole_card_text_size=0.025,
    name_x=0,
    name_y=0,
    name_box_width=0.225,
    name_box_height=0.03,
    name_box_radius=0,
    name_box_color='darkblue',
    name_text_style='bold',
    name_text_color='white',
    name_text_font='sans',
    name_text_size=0.02,
    stack_x=0,
    stack_y=-0.03,
    stack_box_width=0.225,
    stack_box_height=0.03,
    stack_box_radius=0,
    stack_box_color='black',
    stack_text_style='bold',
    stack_text_color='white',
    stack_text_font='sans',
    stack_text_size=0.02,
    previous_action_x=0,
    previous_action_y=-0.06,
    previous_action_box_width=0.225,
    previous_action_box_height=0.03,
    previous_action_box_radius=0,
    previous_action_box_color='darkgray',
    previous_action_text_style='bold',
    previous_action_text_color='black',
    previous_action_text_font='sans',
    previous_action_text_size=0.02,
    club_color='green',
    diamond_color='blue',
    heart_color='red',
    spade_color='black',
    unknown_color='white',
    shifter_timeout=0.25,
    watchdog_timeout=1,
    frame_rate=1000,
)
DEFAULT_ROOT_ROUTINGCONF: str = 'cardroom.routing'
DEFAULT_GAMEMASTER_TIMEOUT: float = 1
_divmod = divmod


def divmod(dividend: int, divisor: int) -> tuple[int, int]:
    quotient: int | float
    remainder: int | float

    match get_decimal_places():
        case 0:
            quotient, remainder = _divmod(int(dividend), int(divisor))
        case math.inf:
            quotient, remainder = dividend / divisor, 0
        case decimal_places:
            assert isinstance(decimal_places, int)

            quotient = (
                floor(dividend / divisor * 10 ** decimal_places)
                / (10 ** decimal_places)
            )
            remainder = dividend - quotient * divisor

    return cast(tuple[int, int], (quotient, remainder))


def parse_value(raw_value: str) -> int:
    value = pokerkit.parse_value(raw_value)

    match get_decimal_places():
        case 0:
            value = int(value)
        case math.inf:
            pass
        case decimal_places:
            assert isinstance(decimal_places, int)

            value = round(value, decimal_places)

    return value


def get_tzinfo() -> ZoneInfo:
    return ZoneInfo(getattr(settings, 'TIME_ZONE', 'UTC'))


def get_divmod() -> Callable[[int, int], tuple[int, int]]:
    return import_string(getattr(settings, 'CARDROOM_DIVMOD', DEFAULT_DIVMOD))


def get_parse_value() -> Callable[[str], int]:
    return import_string(
        getattr(settings, 'CARDROOM_PARSE_VALUE', DEFAULT_PARSE_VALUE),
    )


def get_decimal_places() -> int | float:
    return getattr(settings, 'CARDROOM_DECIMAL_PLACES', DEFAULT_DECIMAL_PLACES)


def get_auth() -> bool:
    return getattr(settings, 'CARDROOM_AUTH', DEFAULT_AUTH)


def get_admin() -> bool:
    return getattr(settings, 'CARDROOM_ADMIN', DEFAULT_ADMIN)


def get_felt() -> bool:
    return getattr(settings, 'CARDROOM_FELT', DEFAULT_FELT)


def get_configuration() -> Configuration:
    return getattr(settings, 'CARDROOM_CONFIGURATION', DEFAULT_CONFIGURATION)


def get_root_routingconf() -> Any:
    return getattr(settings, 'ROOT_ROUTINGCONF', DEFAULT_ROOT_ROUTINGCONF)


def get_gamemaster_timeout() -> Any:
    return getattr(settings, 'GAMEMASTER_TIMEOUT', DEFAULT_GAMEMASTER_TIMEOUT)


def serialize(obj: Any) -> Any:
    if obj is None or isinstance(obj, bytes | str | int | float | bool):
        return obj
    elif is_dataclass(obj):
        return serialize(asdict(obj))
    elif isinstance(obj, Mapping):
        return dict(zip(serialize(obj.keys()), serialize(obj.values())))
    elif isinstance(obj, Iterable):
        return list(map(serialize, obj))
    else:
        raise AssertionError



from rest_framework.request import Request
from rest_framework.serializers import (
    HyperlinkedModelSerializer,
    SerializerMethodField,
)

from cardroom.models import CashGame, HandHistory, Poker, Table
from cardroom.utilities import get_felt


class CardroomHyperlinkedModelSerializer(HyperlinkedModelSerializer):
    @property
    def request(self) -> Request:
        return self.context['request']


class PokerSerializer(CardroomHyperlinkedModelSerializer):
    class Meta:
        model = Poker
        fields = '__all__'


class TableSerializer(CardroomHyperlinkedModelSerializer):
    class Meta:
        model = Table
        fields = '__all__'


class CashGameSerializer(CardroomHyperlinkedModelSerializer):
    frame_url = SerializerMethodField()

    def get_frame_url(self, obj: CashGame) -> str:
        return self.request.build_absolute_uri(obj.get_frame_url())

    websocket_url = SerializerMethodField()

    def get_websocket_url(self, obj: CashGame) -> str:
        url = self.request.build_absolute_uri(obj.get_websocket_url())

        if url.startswith('https:'):
            url = f'wss:{url[6:]}'
        elif url.startswith('http:'):
            url = f'ws:{url[5:]}'
        else:
            raise AssertionError

        return url

    if get_felt():
        felt_url = SerializerMethodField()

        def get_felt_url(self, obj: CashGame) -> str:
            return self.request.build_absolute_uri(obj.get_felt_url())

    class Meta:
        model = CashGame
        fields = '__all__'


class HandHistorySerializer(CardroomHyperlinkedModelSerializer):
    url_field_name = 'url_'
    frames_url = SerializerMethodField()

    def get_frames_url(self, obj: HandHistory) -> str:
        return self.request.build_absolute_uri(obj.get_frames_url())

    if get_felt():
        felt_url = SerializerMethodField()

        def get_felt_url(self, obj: HandHistory) -> str:
            return self.request.build_absolute_uri(obj.get_felt_url())

    class Meta:
        model = HandHistory
        fields = '__all__'



from django.apps import AppConfig


class CardroomConfig(AppConfig):
    default_auto_field = 'django.db.models.BigAutoField'
    name = 'cardroom'
    thread = None

    def ready(self) -> None:
        __import__('cardroom.signals')



from functools import partial
from sys import argv

from django.core.management import execute_from_command_line

main = partial(execute_from_command_line, argv)

if __name__ == '__main__':
    main()  # pragma: no cover



# Generated by Django 4.2.9 on 2024-01-30 06:13

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='HandHistory',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('variant', models.CharField(choices=[('FT', "Fixed-limit Texas hold 'em"), ('NT', "No-limit Texas hold 'em"), ('NS', "No-limit short-deck hold 'em"), ('PO', "Pot-limit Omaha hold 'em"), ('FO/8', "Fixed-limit Omaha hold 'em high/low-split eight-or-better"), ('F7S', 'Fixed-limit seven card stud'), ('F7S/8', 'Fixed-limit seven card stud high/low-split eight-or-better'), ('FR', 'Fixed-limit razz'), ('N2L1D', 'No-limit deuce-to-seven lowball single draw'), ('F2L3D', 'Fixed-limit deuce-to-seven lowball triple draw'), ('FB', 'Fixed-limit badugi')], max_length=255)),
                ('ante_trimming_status', models.BooleanField(default=False)),
                ('antes', models.JSONField()),
                ('blinds_or_straddles', models.JSONField(blank=True, null=True)),
                ('bring_in', models.JSONField(blank=True, null=True)),
                ('small_bet', models.JSONField(blank=True, null=True)),
                ('big_bet', models.JSONField(blank=True, null=True)),
                ('min_bet', models.JSONField(blank=True, null=True)),
                ('starting_stacks', models.JSONField()),
                ('actions', models.JSONField()),
                ('author', models.CharField(blank=True, max_length=255, null=True)),
                ('event', models.CharField(blank=True, max_length=255, null=True)),
                ('url', models.CharField(blank=True, max_length=255, null=True)),
                ('address', models.CharField(blank=True, max_length=255, null=True)),
                ('city', models.CharField(blank=True, max_length=255, null=True)),
                ('region', models.CharField(blank=True, max_length=255, null=True)),
                ('postal_code', models.CharField(blank=True, max_length=255, null=True)),
                ('country', models.CharField(blank=True, max_length=255, null=True)),
                ('time', models.TimeField(blank=True, null=True)),
                ('time_zone', models.CharField(blank=True, max_length=255, null=True)),
                ('day', models.PositiveBigIntegerField(blank=True, null=True)),
                ('month', models.PositiveBigIntegerField(blank=True, null=True)),
                ('year', models.PositiveBigIntegerField(blank=True, null=True)),
                ('hand', models.PositiveBigIntegerField(blank=True, null=True)),
                ('level', models.PositiveBigIntegerField(blank=True, null=True)),
                ('seats', models.JSONField(blank=True, null=True)),
                ('seat_count', models.PositiveBigIntegerField(blank=True, null=True)),
                ('table', models.PositiveBigIntegerField(blank=True, null=True)),
                ('players', models.JSONField(blank=True, null=True)),
                ('finishing_stacks', models.JSONField(blank=True, null=True)),
                ('currency', models.CharField(blank=True, max_length=255, null=True)),
                ('time_limit', models.JSONField(blank=True, null=True)),
                ('time_banks', models.JSONField(blank=True, null=True)),
            ],
            options={
                'verbose_name_plural': 'hand histories',
            },
        ),
        migrations.CreateModel(
            name='Poker',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('variant', models.CharField(choices=[('FT', "Fixed-limit Texas hold 'em"), ('NT', "No-limit Texas hold 'em"), ('NS', "No-limit short-deck hold 'em"), ('PO', "Pot-limit Omaha hold 'em"), ('FO/8', "Fixed-limit Omaha hold 'em high/low-split eight-or-better"), ('F7S', 'Fixed-limit seven card stud'), ('F7S/8', 'Fixed-limit seven card stud high/low-split eight-or-better'), ('FR', 'Fixed-limit razz'), ('N2L1D', 'No-limit deuce-to-seven lowball single draw'), ('F2L3D', 'Fixed-limit deuce-to-seven lowball triple draw'), ('FB', 'Fixed-limit badugi')], max_length=255)),
                ('ante_trimming_status', models.BooleanField(default=False)),
                ('raw_antes', models.JSONField()),
                ('raw_blinds_or_straddles', models.JSONField(blank=True, null=True)),
                ('bring_in', models.JSONField(blank=True, null=True)),
                ('small_bet', models.JSONField(blank=True, null=True)),
                ('big_bet', models.JSONField(blank=True, null=True)),
                ('min_bet', models.JSONField(blank=True, null=True)),
            ],
            options={
                'verbose_name_plural': 'poker',
            },
        ),
        migrations.CreateModel(
            name='Table',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('seat_count', models.PositiveBigIntegerField()),
                ('min_starting_stack', models.JSONField(blank=True, null=True)),
                ('max_starting_stack', models.JSONField(blank=True, null=True)),
                ('game', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to='cardroom.poker')),
            ],
        ),
        migrations.CreateModel(
            name='CashGame',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('time_bank', models.FloatField()),
                ('time_bank_increment', models.FloatField()),
                ('state_construction_timeout', models.FloatField()),
                ('state_destruction_timeout', models.FloatField()),
                ('idle_timeout', models.FloatField()),
                ('standing_pat_timeout', models.FloatField()),
                ('betting_timeout', models.FloatField()),
                ('hole_cards_showing_or_mucking_timeout', models.FloatField()),
                ('table', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to='cardroom.table')),
            ],
            options={
                'abstract': False,
            },
        ),
    ]
